from urllib.parse import quote

from django.contrib.auth import get_user_model
from django.http import HttpResponseRedirect
from django.urls import reverse

from social_core.pipeline import social_auth, user as social_core_user
from social_core.pipeline.user import USER_FIELDS

from .config import UNICEF_EMAIL


def social_details(backend, details, response, *args, **kwargs):
    resp = social_auth.social_details(backend, details, response, *args, **kwargs)
    user = kwargs.get("user")
    if user:
        # here we are preventing messing up between current us and social user
        unauthorized = reverse("unicef_security:unauthorized")
        return HttpResponseRedirect(f"{unauthorized}?eu={user.email}&msgc=alreadyauthenticated")

    resp["details"]["idp"] = response.get("idp")

    if not resp["details"].get("email"):  # pragma: no cover
        resp["details"]["email"] = (
            response.get("email") if response.get("email") else response["signInNames.emailAddress"]
        )
    email = resp["details"].get("email")
    resp["details"]["email"] = email.lower().strip()
    return resp


def get_username(strategy, details, backend, user=None, *args, **kwargs):
    username = details.get("email")

    try:
        get_user_model().objects.get(username=username)

    except get_user_model().DoesNotExist:
        email = quote(username)
        unauthorized = reverse("unicef_security:unauthorized")
        return HttpResponseRedirect(f"{unauthorized}?eu={email}&msgc=nouser")

    return {"username": details.get("email")}


def create_unicef_user(strategy, details, backend, user=None, *args, **kwargs):
    """Override create_user, to create only UNICEF users."""
    if user:
        return {"is_new": False}

    fields = {name: kwargs.get(name, details.get(name)) for name in backend.setting("USER_FIELDS", USER_FIELDS)}

    if not (fields and details.get("email", "").endswith(UNICEF_EMAIL)):
        return None

    response = kwargs.get("response")
    if response:
        email = response.get("email") or response.get("signInNames.emailAddress")
        if email and not email.endswith(UNICEF_EMAIL):
            return None

    return {"is_new": True, "user": strategy.create_user(**fields)}


def user_details(strategy, details, backend, user=None, *args, **kwargs):
    """Update the user see what the property to map by is here updates_available = False."""
    if user:
        user.username = user.email
        user.first_name = details["first_name"]
        user.last_name = details["last_name"]
        user.set_unusable_password()
        user.save()

    return social_core_user.user_details(strategy, details, backend, user, *args, **kwargs)



from typing import Iterable

from django.conf import settings

from social_core.backends.azuread_b2c import AzureADB2COAuth2
from social_core.exceptions import AuthCanceled, AuthMissingParameter
from social_django.middleware import SocialAuthExceptionMiddleware

from . import config


class UNICEFSocialAuthExceptionMiddleware(SocialAuthExceptionMiddleware):
    """Middleware to ignore Forgot Password Exceptions."""

    def process_exception(self, request, exception):
        if exception in [AuthCanceled, AuthMissingParameter]:
            return self.get_redirect_uri(request, exception)
        return super().process_exception(request, exception)  # pragma: no cover

    def get_redirect_uri(self, request, exception):
        strategy = getattr(request, "social_strategy", None)
        error = request.GET.get("error", None)

        # This is what we should expect:
        # ['AADB2C90118: The user has forgotten their password.\r\n
        # Correlation ID: 7e8c3cf9-2fa7-47c7-8924-a1ea91137ba9\r\n
        # Timestamp: 2018-11-13 11:37:56Z\r\n']
        error_description = request.GET.get("error_description", None)
        if (
            error == "access_denied" and isinstance(error_description, Iterable) and "AADB2C90118" in error_description
        ):  # pragma: no cover
            auth_class = AzureADB2COAuth2()
            redirect_home = auth_class.get_redirect_uri()
            reset_policy = config.AZURE_RESET_POLICY
            return "".join(
                [
                    auth_class.base_url,
                    "/oauth2/v2.0/",
                    f"authorize?p={reset_policy}",
                    f"&client_id={strategy.setting('KEY')}",
                    f"&nonce=defaultNonce&redirect_uri={redirect_home}",
                    "&scope=openid+email&response_type=code",
                ]
            )

        # TODO: In case of password reset the state can't be verified figure out a way to log the user in after reset
        if error is None:  # pragma: no cover
            return settings.LOGIN_URL

        strategy = getattr(request, "social_strategy", None)
        return strategy.setting("LOGIN_ERROR_URL") + "?msgc=loginerror"



from django.urls import re_path

from unicef_security.views import UnauthorizedView, UNICEFLogoutView

app_name = "unicef_security"

urlpatterns = [
    re_path(r"^unicef-logout/$", UNICEFLogoutView.as_view(), name="unicef-logout"),
    re_path(r"^unauthorized/$", UnauthorizedView.as_view(), name="unauthorized"),
]



import os
from typing import Any

from django.http import HttpRequest


def current_state(request: HttpRequest) -> dict[str, Any]:
    return {
        "app": {
            "version": os.environ.get("VERSION", ""),
            "build_date": os.environ.get("BUILD_DATE", ""),
            "commit": os.environ.get("GIT_SHA", "-"),
        },
    }



# Generated by Django 3.2 on 2022-04-06 23:07

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("unicef_security", "0002_auto_20201028_0154"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="user",
            options={"verbose_name": "user", "verbose_name_plural": "users"},
        ),
        migrations.DeleteModel(
            name="BusinessArea",
        ),
        migrations.DeleteModel(
            name="Region",
        ),
    ]



import logging
from typing import Any

from django import forms
from django.contrib import messages
from django.contrib.admin import SimpleListFilter
from django.contrib.auth.admin import UserAdmin
from django.http import HttpRequest
from django.template.response import TemplateResponse
from django.utils.translation import gettext_lazy as _

from admin_extra_buttons.decorators import button
from admin_extra_buttons.mixins import ExtraButtonsMixin

from .config import UNICEF_EMAIL
from .graph import default_group, Synchronizer, SyncResult

logger = logging.getLogger(__name__)


def is_superuser(request, *args, **kwargs):  # pragma: no cover
    return request.user.is_superuser


class LoadUsersForm(forms.Form):
    emails = forms.CharField(widget=forms.Textarea)


class UNICEFUserFilter(SimpleListFilter):
    title = "UNICEF user filter"
    parameter_name = "email"

    def lookups(self, request, model_admin):
        return (
            (
                "unicef",
                _(
                    "UNICEF",
                ),
            ),
            (
                "external",
                _(
                    "External",
                ),
            ),
        )

    def queryset(self, request, queryset):
        if self.value():
            if "unicef" in self.value():
                queryset = queryset.filter(email__endswith=UNICEF_EMAIL)
            elif "external" in self.value():
                queryset = queryset.exclude(email__endswith=UNICEF_EMAIL)
        return queryset


class UserAdminPlus(ExtraButtonsMixin, UserAdmin):
    list_display = [
        "username",
        "first_name",
        "last_name",
        "display_name",
        "email",
        "is_staff",
        "is_active",
        "is_superuser",
        "is_linked",
        "last_login",
    ]
    list_filter = ["is_superuser", "is_staff", "is_active", UNICEFUserFilter]
    search_fields = ["username", "display_name", "first_name", "last_name"]
    fieldsets = (
        (None, {"fields": (("username", "azure_id"), "password")}),
        (
            _("Personal info"),
            {
                "fields": (
                    (
                        "first_name",
                        "last_name",
                    ),
                    ("email", "display_name"),
                    ("job_title",),
                )
            },
        ),
        (
            _("Important dates"),
            {
                "classes": ["collapse"],
                "fields": (
                    "last_login",
                    "date_joined",
                    "last_modify_date",
                ),
            },
        ),
    )

    add_fieldsets = (
        (
            None,
            {
                "classes": ("wide",),
                "fields": ("username", "password1", "password2", "email"),
            },
        ),
    )

    extra_fieldsets = (
        _("Permissions"),
        {
            "fields": (
                ("password",),
                (
                    "is_active",
                    "is_staff",
                    "is_superuser",
                ),
                ("groups",),
            ),
        },
    )
    readonly_fields = ("azure_id", "job_title", "display_name", "last_modify_date")

    def get_fieldsets(self, request: HttpRequest, obj: Any | None = None) -> tuple:
        if not obj:
            return self.add_fieldsets
        fieldsets = super().get_fieldsets(request, obj)
        if request.user.is_superuser:
            fieldsets = fieldsets + self.extra_fieldsets
        return fieldsets

    def is_linked(self, obj):
        return bool(obj.azure_id)

    is_linked.boolean = True

    @button(label="Sync")
    def sync_user(self, request, pk):
        obj = self.get_object(request, pk)
        try:
            synchronizer = Synchronizer()
            synchronizer.sync_user(obj)
        except ValueError as e:  # pragma: no cover
            self.message_user(request, str(e), messages.ERROR)

        self.message_user(request, "User synchronized")

    @button(label="Link user", permissions=is_superuser)
    def link_user_data(self, request, pk):
        opts = self.model._meta
        ctx = {
            "opts": opts,
            "app_label": "security",
            "change": True,
            "is_popup": False,
            "save_as": False,
            "has_delete_permission": False,
            "has_add_permission": False,
            "has_change_permission": True,
        }
        obj = self.get_object(request, pk)
        synchronizer = Synchronizer()
        try:
            if request.method == "POST":
                if request.POST.get("selection"):
                    data = synchronizer.get_user(request.POST.get("selection"))
                    synchronizer.sync_user(obj, data["id"])
                    self.message_user(request, "User linked")
                    return None
                ctx["message"] = "Select one entry to link"

            data = synchronizer.search_users(obj)
            ctx["data"] = data
            return TemplateResponse(request, "admin/link_user.html", ctx)

        except ValueError as e:  # pragma: no cover
            self.message_user(request, str(e), messages.ERROR)

    @button()
    def load(self, request):
        opts = self.model._meta
        ctx = {
            "opts": opts,
            "app_label": "security",
            "change": True,
            "is_popup": False,
            "save_as": False,
            "has_delete_permission": False,
            "has_add_permission": False,
            "has_change_permission": True,
        }
        if request.method == "POST":
            form = LoadUsersForm(request.POST)
            if form.is_valid():
                synchronizer = Synchronizer()
                emails = form.cleaned_data["emails"].split()
                total_results = SyncResult()
                for email in emails:
                    result = synchronizer.fetch_users("startswith(mail,'%s')" % email, callback=default_group)
                    total_results += result
                self.message_user(
                    request,
                    f"{len(total_results.created)} users have been created, "
                    f"{len(total_results.updated)} updated. "
                    f"{len(total_results.skipped)} invalid entries found.",
                )
        else:
            form = LoadUsersForm()
        ctx["form"] = form
        return TemplateResponse(request, "admin/load_users.html", ctx)

    @button(permissions=is_superuser)
    def ad(self, request, pk):
        obj = self.get_object(request, pk)
        context = dict
        try:
            synchronizer = Synchronizer()
            context = synchronizer.get_user(obj.username)
        except ValueError as e:  # pragma: no cover
            self.message_user(request, str(e), messages.ERROR)

        return TemplateResponse(request, "admin/ad.html", {"ctx": context, "opts": self.model._meta})



import logging

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group

import requests
from constance import config as constance
from social_django.models import UserSocialAuth

from unicef_security.config import GRAPH_CLIENT_ID, GRAPH_CLIENT_SECRET

from . import config

AZURE_GRAPH_API_TOKEN_CACHE_KEY = "azure_graph_api_token_cache_key"  # noqa
AZURE_GRAPH_DELTA_LINK_KEY = "azure_graph_delta_link_key"

logger = logging.getLogger(__name__)

DJANGOUSERMAP = {
    "_pk": ["username"],
    "username": "userPrincipalName",
    "email": "mail",
    "azure_id": "id",
    "job_title": "jobTitle",
    "display_name": "displayName",
    "first_name": "givenName",
    "last_name": "surname",
}

ADMIN_EMAILS = [i[1] for i in settings.ADMINS]


def default_group(**kwargs):
    is_new = kwargs.get("is_new", False)
    user = kwargs.get("user")
    if is_new:
        if user.email in ADMIN_EMAILS:
            user.is_staff = True
            user.is_superuser = True
            user.save()
        elif group_name := constance.DEFAULT_GROUP:
            group = Group.objects.filter(name=group_name).first()
            if group:
                user.groups.add(group)


def get_unicef_user(backend, details, response, *args, **kwargs):
    User = get_user_model()  # noqa
    if details.get("email"):
        filters = {"email": details["email"]}
    elif details.get("unique_name"):
        filters = {"username": details["unique_name"]}
    elif details.get("username"):
        filters = {"username": details["username"]}

    try:
        user = User.objects.get(**filters)
        social = user.social_auth.get()
        user.social_user = social
        created = False
    except (User.DoesNotExist, UserSocialAuth.DoesNotExist):
        for k, v in response.items():
            if k in ["email", "family_name", "unique_name"]:
                details[k] = v
        try:
            sync = Synchronizer()
            data = sync.get_user(details["email"])

            for k, v in data.items():
                details[k] = v

        except (ValueError, KeyError) as e:  # pragma: no cover
            logger.error(e)

        user, created = User.objects.get_or_create(
            username=details["unique_name"],
            defaults={
                "first_name": details.get("givenName", ""),
                "last_name": details.get("surname", ""),
                "job_title": details.get("jobTitle", ""),
                "display_name": details.get("displayName", details["unique_name"]),
                "email": details.get("email", ""),
                "azure_id": details.get("id"),
            },
        )
        social, _ = UserSocialAuth.objects.get_or_create(user=user, provider=backend.name, uid=user.username)
        user.social_user = social
    return {"user": user, "social": social, "uid": details.get("id"), "is_new": created}


class SyncResult:
    def __init__(self, keep_records=False):
        """Class to synchronize against external providers.

        :param keep_records: if True keep track of record instances
        """
        self.created = []
        self.updated = []
        self.skipped = []
        self.keep_records = keep_records

    def log(self, *result):
        if len(result) == 2:
            obj = {True: result[0], False: result[0].pk}[self.keep_records]
            if result[1]:
                self.created.append(obj)
            else:
                self.updated.append(obj)
        else:
            self.skipped.append(result)

    def __add__(self, other):
        if isinstance(other, SyncResult):
            ret = SyncResult()
            ret.created = self.created + other.created
            ret.updated = self.updated + other.updated
            ret.skipped = self.skipped + other.skipped
            return ret
        raise ValueError("Cannot add %s to SyncResult object" % type(other))

    def __repr__(self):
        return f"<SyncResult: {len(self.created)} {len(self.updated)} {len(self.skipped)}>"

    def __eq__(self, other):
        if isinstance(other, SyncResult):
            return self.created == other.created and self.updated == other.updated and self.skipped == other.skipped
        return False


NotSet = object()


class Synchronizer:
    def __init__(self, user_model=None, mapping=None, echo=None, identifier=None, secret=None):
        self.id = identifier or GRAPH_CLIENT_ID
        self.secret = secret or GRAPH_CLIENT_SECRET

        self.user_model = user_model or get_user_model()
        self.field_map = dict(mapping or DJANGOUSERMAP)
        self.user_pk_fields = self.field_map.pop("_pk")
        self._baseurl = f"{config.AZURE_GRAPH_API_BASE_URL}/{config.AZURE_GRAPH_API_VERSION}/users"
        self.startUrl = "%s/delta" % self._baseurl
        self.access_token = self.get_token()
        self.next_link = None
        self._delta_link = ""
        self.echo = echo or (lambda lmn: True)

    def get_token(self):
        if not self.id and self.secret:  # pragma: no cover
            raise ValueError("Configure AZURE_CLIENT_ID and/or AZURE_CLIENT_SECRET")
        post_dict = {
            "grant_type": "client_credentials",
            "client_id": self.id,
            "client_secret": self.secret,
            "resource": config.AZURE_GRAPH_API_BASE_URL,
        }
        response = requests.post(f"{config.AZURE_URL}/unicef.org/oauth2/token", post_dict, timeout=60)
        if response.status_code != 200:  # pragma: no cover
            logger.error(f"Unable to fetch token from Azure. {response.status_code} {response.content}")
            raise BaseException(f"Error during token retrieval: {response.status_code} {response.content}")
        jresponse = response.json()
        return jresponse["access_token"]

    @property
    def delta_link(self):
        return self._delta_link

    @delta_link.setter
    def delta_link(self, value):
        self._delta_link = value

    def get_page(self, url, single=False):
        while True:
            headers = {"Authorization": f"Bearer {self.get_token()}"}
            try:
                response = requests.get(url, headers=headers, timeout=60)
                if response.status_code == 401:  # pragma: no cover
                    data = response.json()
                    if data["error"]["message"] == "Access token has expired.":
                        continue
                    raise ConnectionError(f"400: Error processing the response {response.content}")

                if response.status_code != 200:  # pragma: no cover
                    raise ConnectionError(
                        f"Code {response.status_code}. Error processing the response {response.content}"
                    )
                break
            except ConnectionError as e:  # pragma: no cover
                logger.exception(e)
                raise

        jresponse = response.json()
        self.next_link = jresponse.get("@odata.nextLink", None)
        self.delta_link = jresponse.get("@odata.deltaLink", None)
        if single:
            return jresponse
        return jresponse.get("value", [])

    def __iter__(self):
        values = self.get_page(self.startUrl)
        pages = 1
        while True:
            try:
                yield values.pop()
            except IndexError:
                if not self.next_link:
                    logger.debug(f"All pages  fetched. deltaLink: {self.delta_link}")
                    break
                values = self.get_page(self.next_link)
                logger.debug(f"fetched page {pages}")
                pages += 1
            except GeneratorExit as e:
                logger.exception(e)
                break

    def get_record(self, user_info: dict) -> (dict, dict):
        data = {fieldname: user_info.get(mapped_name, "") for fieldname, mapped_name in self.field_map.items()}
        pk = {fieldname: data.pop(fieldname) for fieldname in self.user_pk_fields}
        return pk, data

    def fetch_users(self, filter_params, max_records=None, callback=None):
        self.startUrl = "%s?$filter=%s" % (self._baseurl, filter_params)
        return self.synchronize(max_records=max_records, callback=callback)

    def search_users(self, record):
        url = "%s?$filter=" % self._baseurl
        filters = []
        field_map = {
            "email": "mail eq '{value}'",
            "last_name": "surname eq '{value}'",
            "first_name": "givenName eq '{value}'",
        }

        for field, filter_template in field_map.items():
            value = getattr(record, field, None)
            if value:
                filters.append(filter_template.format(value=value))

        page = self.get_page(url + " or ".join(filters), single=True)
        return page["value"]

    def filter_users_by_email(self, email):
        """Filter users by email.

        https://graph.microsoft.com/v1.0/users?$filter=mail eq 'sapostolico@unicef.org'
        """
        url = "%s?$filter=mail eq '%s'" % (self._baseurl, email)
        page = self.get_page(url, single=True)
        return page["value"]

    def get_user(self, username):
        url = "%s/%s" % (self._baseurl, username)
        return self.get_page(url, single=True)

    def sync_user(self, user, azure_id=None):
        if not (azure_id or user.azure_id):
            raise ValueError("Cannot sync user without azure_id")
        url = "%s/%s" % (self._baseurl, azure_id or user.azure_id)
        user_info = self.get_page(url, single=True)
        pk, values = self.get_record(user_info)
        user, __ = self.user_model.objects.update_or_create(**pk, defaults=values)
        return user

    def resume(self, *, delta_link=None, max_records=None):
        if delta_link:
            self.startUrl = delta_link
        return self.synchronize(max_records)

    def is_valid(self, user_info):
        return user_info.get("email")

    def synchronize(self, max_records=None, callback=None):
        logger.debug("Start Azure user synchronization")
        results = SyncResult()
        try:
            for i, user_info in enumerate(iter(self)):
                pk, values = self.get_record(user_info)
                if self.is_valid(values):
                    user, created = self.user_model.objects.update_or_create(**pk, defaults=values)
                    if callback:
                        callback(user=user, is_new=created)
                    self.echo([user, created])
                    results.log(user, created)
                else:
                    results.log(user_info)
                if max_records and i > max_records:
                    break
        except BaseException as e:  # pragma: no cover
            logger.exception(e)
            raise
        logger.debug(f"End Azure user synchronization: {results}")
        return results



import os

from jwt import decode, DecodeError, ExpiredSignatureError, get_unverified_header
from social_core.backends.azuread_tenant import AzureADTenantOAuth2
from social_core.exceptions import AuthTokenError


class UNICEFAzureADTenantOAuth2Ext(AzureADTenantOAuth2):
    name = "unicef-azuread-tenant-oauth2"

    def user_data(self, access_token, *args, **kwargs):
        response = kwargs.get("response")
        id_token = response.get("id_token")

        # get key id and algorithm
        key_id = get_unverified_header(id_token)["kid"]
        key = ""
        verify = os.environ.get("OAUTH2_VERIFY", "")
        try:
            # retrieve certificate for key_id
            if verify:  # pragma: no cover
                certificate = self.get_certificate(key_id)
                key = certificate.public_key()

            options = {"verify_signature": verify}
            return decode(
                id_token,
                key=key,
                algorithms=["RS256"],
                audience=self.setting("KEY"),
                options=options,
            )
        except (DecodeError, ExpiredSignatureError) as error:  # pragma: no cover
            raise AuthTokenError(self, error)



# Generated by Django 4.2.1 on 2023-11-09 11:12

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("unicef_security", "0003_auto_20220406_2307"),
    ]

    operations = [
        migrations.DeleteModel(
            name="User",
        ),
    ]



from django.conf import settings
from django.template import Library
from django.urls import reverse

register = Library()


@register.simple_tag()
def ad_backend():
    return reverse(
        "social:begin",
        args=[getattr(settings, "SOCIAL_AUTH_BACKEND_NAME", "azuread-tenant-oauth2")],
    )


@register.simple_tag
def settings_value(name):
    return getattr(settings, name, "")



from django.contrib.auth.models import AbstractUser
from django.db import models
from django.utils.functional import cached_property

app_label = "unicef_security"


class TimeStampedModel:
    last_modify_date = models.DateTimeField(editable=False, blank=True, auto_now_add=True, auto_now=True)


class SecurityMixin(models.Model):
    azure_id = models.UUIDField(blank=True, unique=True, null=True)
    job_title = models.CharField(max_length=100, null=True, blank=True)
    display_name = models.CharField(max_length=100, null=True, blank=True)

    class Meta(AbstractUser.Meta):
        abstract = True
        app_label = "unicef_security"

    def save(self, *args, **kwargs):
        if not self.display_name:
            self.display_name = self.label
        super().save(*args, **kwargs)

    @cached_property
    def label(self):
        if self.display_name:
            return self.display_name
        if self.first_name and self.last_name:
            return f"{self.first_name} {self.last_name}:"
        if self.first_name:
            return self.first_name
        return self.username
